import os, time, math, datetime
os.system('cls')
import numpy as np

import jdatetime
from survey_stats.data_process import Data, Date
from survey_stats.functions import subsets, list_to_str, match_str, match_in

import shutdown

# shutdown.shutdown(10)





from __future__ import annotations
from typing import Union
import numpy as np
import warnings
warnings.filterwarnings("ignore")
import jdatetime, datetime, time, math
import random

def number_of_digits(n:int, decimals:float=0)->int:
    try:
        return len(str(int(n*10**decimals)/10**decimals))
    except:
        return 5

def subsets(a_set:list, deep:int=-1, randomly:bool=False)->list:
    '''
    deep is a level base on len of subsets.\n
    randomly == True to select randomly subsets per level as
    '''
    subs = []
    if deep==-1:
        if len(a_set) == 0:
            return [[]]
        else:
            s = subsets(a_set[:-1])
            return s + [i + [a_set[-1]] for i in s]
    elif not randomly:
        if a_set == []:
            return []
        if deep == 0:
            return a_set
        else:
            subs = [a_set[1:]]
            for i in range(1, len(a_set)-1):
                subs.append(a_set[:i]+a_set[i+1:])
            subs.append(a_set[:-1])
            if deep > 1:
                for sub in subs[1:]:
                    subs.extend(subsets(sub, deep-1, randomly)[1:])
            subs = [a_set] + subs
    else:
        if a_set == []:
            return []
        if deep == 0:
            return a_set
        else:
            subs = [a_set[1:]]
            for i in range(1, len(a_set)-1):
                subs.append(a_set[:i]+a_set[i+1:])
            subs.append(a_set[:-1])
            i = 1
            while  deep >= 2:
                for sub in subs[i:]:
                    subsubs = subsets(sub, 1, True)
                    subs.append(subsubs[random.randint(1,len(subsubs)-1)])
                    i += 1
                deep -= 1
            subs = [a_set] + subs
    subs.sort(key=lambda x:len(x), reverse=True)
    return subs

def to_number(x:str, skip_chars:bool=True)->float:
    j, has_seprator = '', False
    for i in x:
        try:
            if i=='.' and not has_seprator:
                j += i
                has_seprator = True
            elif not i=='.':
                j += str(int(i))
            else:
                if not skip_chars:
                    return np.nan
        except:
            if not skip_chars:
                return np.nan
    if j=='':
        return np.nan
    else:
        return float(j)

def remove_number(x:str)->str:
    j, has_seprator = '', False
    for i in x:
        if not i.isnumeric() and (i!='.' or has_seprator):
            j+=i
        elif not has_seprator and i=='.':
            has_seprator = True
    return j

def split_str_number(x:str)->tuple[str, float]:
    j_str, j_int, has_seprator = '', '', False
    for i in x:
        if i == '.' and not has_seprator:
            j_int += i
            has_seprator = True
        elif not i == '.' and i.isnumeric():
            j_int += str(int(i))
        elif  i=='.' or (i!='.' or  has_seprator):
            j_str += i
    if j_int == '':
        j_int = np.nan
    else:
        j_int = float(j_int)
    return (j_str, j_int)

def check_condition(value:Union[float, str], sign:str, criteria:Union[float,str])->bool:
    err = f"Error! '{sign}' is not use for comparing '{criteria}' and '{value}'."
    if sign == '>=':
        if type(criteria)==float or type(criteria)==int:
            res = value >= criteria
        else:
            raise ValueError(err)
    elif sign == '<=':
        if type(criteria)==float or type(criteria)==int:
            res = value <= criteria
        else:
            raise ValueError(err)
    elif sign == '!=':
        res = value != criteria
    elif sign == '>':
        if type(criteria)==float or type(criteria)==int:
            res = value > criteria
        else:
            raise ValueError(err)
    elif sign == '<':
        if type(criteria)==float or type(criteria)==int:
            res = value < criteria
        else:
            raise ValueError(err)
    elif sign == '=':
        res = value == criteria
    else:
        err = f"Error! '{sign}' must be one of the symbols " + \
                            "'=','!=','>','>=','<','<=', or '!='."
        raise ValueError(err)
    return res

def seconds_to_days_hms(seconds:float)->str:
    # decs = seconds - int(seconds)
    days = int(seconds // (24*60*60))
    seconds_remind = seconds-days*24*60*60
    hours = int(seconds_remind // (60*60))
    seconds_remind = seconds_remind - hours*60*60
    minutes = int(seconds_remind // 60)
    seconds_remind = seconds_remind - minutes*60 
    res = ''
    if days != 0:
        res += f'{days}days '
    if hours != 0:
        res += f'{hours}h '
    if minutes != 0:
        res += f'{minutes}m '
    if seconds_remind != 0:
        res += f'{seconds_remind:.0f}s'
    return res

def is_numeric(x:Union[str, int, float, bool])->bool:
    if isinstance(x, int) or isinstance(x, float):
        return True
    return False

def is_numeric_str(text:str)->bool:
    text.replace(' ', '')
    if text == '':
        return False
    elif len(text.split('e'))==2:
        if is_numeric_str(text.split('e')[0]) and is_numeric_str(text.split('e')[1]):
            return True
    elif len(text.split('E'))==2:
        if is_numeric_str(text.split('E')[0]) and is_numeric_str(text.split('E')[1]):
            return True
    elif text[0] == '-':
        text = text[1:]
    try:
        if len(text.split('.'))==2:
            return min([i.isdigit() for i in text.split('.') if i!=''])
        else:
            return False
    except:
        return False

def to_float(text:str)->float:
    text.replace(' ', '')
    if len(text.split('e'))==2:
        if is_numeric_str(text.split('e')[0]) and is_numeric_str(text.split('e')[1]):
            return float(text.split('e')[0]) * 10** float(text.split('e')[1])
    if is_numeric_str(text):
        return float(text)
    else:
        return np.nan

def days_of_month(year:int, month:int)->int:
    if 0<month <7:
        return 31
    elif month<12:
        return 30
    elif month==12 and ((1244<int(year)<=1342 and int(year)%33 in [1,5,9,13,17,21,26,30])
                    or (1342<int(year)<=1472 and int(year)%33 in [1,5,9,13,17,22,26,30])):
        return 30
    elif month==12:
        return 29
    else:
        raise ValueError(f"Error month of '{month}' must between 1 to 12.")

def date_to_standard(date:str|int, date_type:str='daily')->str:
    if date_type in ['daily', 'weekly', 'monthly', 'seasonal']:
        if not('/' in date or '-' in date):
            raise ValueError(f"Error! '{date}' must has '/' or '-' as seprator.")
    if date_type in ['daily', 'weekly']:
        splits = date.split('/') if '/' in date else date.split('-')
        if len(splits)!=3:
            raise ValueError(f"Error! the number of splits must be 3.")
        y, m, d = splits
        max_days = days_of_month(int(y), int(m))
        if int(d)<1 or int(d)>max_days:
            raise ValueError(f"Error days of in {y}-{m} must between 1 to {max_days}.")
        m = '0'+m if len(m)==1 else m
        d = '0'+d if len(d)==1 else d
        return f'{y}-{m}-{d}'
    elif date_type == 'monthly':
        splits = date.split('/') if '/' in date else date.split('-')
        if len(splits)!=2:
            raise ValueError(f"Error! the number of splits must be 2.")
        y, m = splits
        m = '0'+m if len(m)==1 else m
        return f'{y}-{m}'
    elif date_type == 'seasonal':
        splits = date.split('/') if '/' in date else date.split('-')
        if len(splits)!=2:
            raise ValueError(f"Error! the number of splits must be 2.")
        y, s = splits
        return f'{y}-{s}'
    elif date_type == 'annual':
        try:
            return int(date)
        except:
            raise ValueError(f"Error! years must be 'int'.")

def xls_read(content:str)->list[list]:
    content = content.replace('<td>', '<th>').replace('</td>', '</th>')
    if not ('<thead>' in content and '</thead>'):
        raise ValueError(f"Error! content don't include 'thead' tag.")
    if not ('<tbody>' in content and '</tbody>'):
        raise ValueError(f"Error! content don't include 'tbody' tag.")
    res = []
    row = []
    headers = content.split('<thead>')[1].split('</thead>')[0].split('<th>')
    for header in headers[1:]:
        row.append(header.split('</th>')[0].strip())
    res.append(row)

    rows = content.split('<tbody>')[1].split('</tbody>')[0].split('<tr>')
    for row in rows[1:]:
        cols = row.split('<th>')
        row = []
        for col in cols[1:]:
            row.append(col.split('</th>')[0].strip())
        res.append(row)
    return res

def to_formated_str(value:str | int | float, decimals:int=4)->str:
    try:
        if value != 0 and (abs(value)<10**(-decimals) or abs(value)>=10**9):
            res = f'{value:,.{decimals}e}'
        else:
            try:
                decimals = min(decimals, len(str(value).split('.')[1]))
            except:
                decimals = 0
            res = f'{value:,.{decimals}f}'
        return res
    except:
        return str(value)

def number_to_jdate(num:int, zero_date:str='1278-10-09')->str:
    if '-' in zero_date and len(zero_date.split('-')) == 3:
        y, m, d = zero_date.split('-')
        sep = '-'
    elif '/' in zero_date and len(zero_date.split('/')) == 3:
        y, m, d = zero_date.split('/')
        sep = '/'
    else:
        raise ValueError(f"Error! seprator of date must be - or /")
    date = jdatetime.datetime(int(y), int(m), int(d)) + jdatetime.timedelta(num)
    return date.strftime(f'%Y{sep}%m{sep}%d')

def number_to_date(num: int, zero_date: str = '1900-01-01')->str:
    if '-' in zero_date and len(zero_date.split('-')) == 3:
        y, m, d = zero_date.split('-')
        sep = '-'
    elif '/' in zero_date and len(zero_date.split('/')) == 3:
        y, m, d = zero_date.split('/')
        sep = '/'
    else:
        raise ValueError(f"Error! seprator of date must be - or /")
    date = datetime.datetime(int(y), int(m), int(d)) + \
        datetime.timedelta(num-2)
    return date.strftime(f'%Y{sep}%m{sep}%d')

def is_nan(value:any, is_number:bool=False)->bool:
    if is_numeric(value):
        if np.isnan(value):
            return True
    elif is_number:
        return True
    return False

def str_to_jdate(date_str:str)->jdatetime.date:
    seprator = '-' if '-' in date_str else '/' if '/' in date_str else ''
    if seprator == '':
        raise ValueError(f"Error! seprator must be '-' or '/'. {date_str} is not standard")
    splits = date_str.split(seprator)
    if len(splits) != 3:
        raise ValueError(
            f"Error! part of date must be 3 (year, month, day). {date_str} is not standard")
    else:
        try:
            y, m, d = [int(x) for x in splits]
        except:
            raise ValueError(
                f"Error! part of date must be integer (year, month, day). {date_str} is not standard")
        return jdatetime.date(y, m, d)

def to_days_later(date_str:str, days:int=1)->str:
    s = '-' if '-' in date_str else '/' if '/' in date_str else ''
    jdate = str_to_jdate(date_str)
    next_date = jdate + jdatetime.timedelta(days)
    return next_date.strftime(f'%Y{s}%m{s}%d')

def to_days_ago(date_str: str, days: int = 1) -> str:
    s = '-' if '-' in date_str else '/' if '/' in date_str else ''
    jdate = str_to_jdate(date_str)
    next_date = jdate - jdatetime.timedelta(days)
    return next_date.strftime(f'%Y{s}%m{s}%d')

def to_weekend(date_str: str) -> str:
    s = '-' if '-' in date_str else '/' if '/' in date_str else ''
    jdate = str_to_jdate(date_str)
    days = 6-jdate.weekday()
    next_date = jdate + jdatetime.timedelta(days)
    return next_date.strftime(f'%Y{s}%m{s}%d')

def today()->str:
    return jdatetime.date.today().strftime(f'%Y-%m-%d')

def progress(start_time:int, iteration:int, iterations:int, len_previous_log:int,
             state:str='', cut_state:bool=True, indent:int = 5):
    elapsed = time.perf_counter() - start_time
    remaining = elapsed * (iterations/(iteration+1) - 1)
    elapsed = seconds_to_days_hms(elapsed)
    remaining = seconds_to_days_hms(remaining)
    log = ' '*indent 
    if state != '':
        if not isinstance(state, str):
            state = str(state)
        if cut_state:
            state = state[:15].ljust(15)
        log += f'{state}-'
    log += f'{iteration+1} of {iterations} ({(iteration+1)/iterations*100:.2f}%). elapsed: {elapsed}. remaining: {remaining}.'
    log += ' '*(max(len_previous_log, len(log))-(len_previous_log:=len(log)))
    if iteration == iterations-1:
        print(log)
    else:
        print(log, end='\r')
    return len_previous_log

def clear_memory():
    for var in dir():
        if not var.startswith('__'):
            del globals()[var]

def match_str(a:str, b:str)->bool:
    if isinstance(a, str) and isinstance(b, str):
        a_s, b_s = a.split('*'), b.split('*')
        if len(a_s)==1:
            if len(b_s)==1:
                return a==b
            else:
                if b_s[0]=='':
                    if b_s[1] in a:
                        if len(b_s[1:])>1:
                            return match_str(a[a.find(b_s[1]):], '*'.join(b_s[1:]))
                        else:
                            return a[a.find(b_s[1]):] == b_s[1]
                    else:
                        return False
                elif b_s[-1]=='':
                    if b_s[-2][::-1] in a[::-1]:
                        if len(b_s[:-1])>1:
                            return match_str(a[:len(a)-a[::-1].find(b_s[-2][::-1])], '*'.join(b_s[:-1]))
                        else:
                            return a[:len(a)-a[::-1].find(b_s[-2][::-1])] ==  b_s[0]
                    else:
                        return False
                else:
                    if a[:len(b_s[0])]!=b_s[0] or a[-len(b_s[-1]):]!=b_s[-1]:
                        return False
                    n = len(b_s[0])
                    for w in b_s[1:-1]:
                        if a[n:].find(w) == -1:
                            return False
                    return True
        else:
            if len(b_s)==1:
                return match_str(b, a)
            else:
                if len(a_s)<len(b_s):
                    return match_str(''.join(a_s), '*'.join(b_s))
                elif len(a_s)>len(b_s):
                    return match_str('*'.join(a_s), ''.join(b_s))
                else:
                    return min([a_s[i] in b_s[i] if len(a_s[i])<=len(b_s[i]) else b_s[i] in a_s[i] for i in range(len(a_s))])
    else:
        return a==b

def match_in(a:str, b:list)->bool:
    for x in b:
        if match_str(a, x):
            return x
    return False

def list_to_str(a_list:list, min_lenght:int=50, indent:int=10):
    res = []
    for j, element in enumerate(a_list):
        if j==0:
            line = ' '*indent+f'{element}'
        elif len(line)<min_lenght:
            line += f', {element}'
        else:
            res.append(line)
            line = ' '*indent+f'{element}'
    if res ==[]:
        res.append(line)
    return res

def quantile(x:list, n:int=10)->list:
    a = sorted([y for y in x if not is_nan(y)])
    thre = [a[int(d*len(a)/n)] for d in range(1,n)]
    res = []
    for y in x:
        if is_nan(y):
            res.append(math.nan)
        else:
            k = 1
            if y<thre[0]:
                res.append(k)
            else:
                for j,d in enumerate(thre):
                    if j>0:
                        k += 1
                        if thre[j-1]<=y<d:
                            res.append(k)
                            break
                else:
                    if d<=y:
                        res.append(k+1)
    return res

def print_back(text:str, back:int=0):
    if back==0:
        print(text)
    else:
        print(f"\033[{back}A{text}"+'\n'*(back-1))



"""
C struct definitions credit

Author: LucasG (https://github.com/lucasg)
Source: http://stackoverflow.com/questions/13564851/generate-keyboard-events
"""

import time
import ctypes

# Import the SendInput object
SendInput = ctypes.windll.user32.SendInput

# C struct redefinitions
PUL = ctypes.POINTER(ctypes.c_ulong)

class KeyBdInput(ctypes.Structure):
    _fields_ = [
        ("wVk", ctypes.c_ushort),
        ("wScan", ctypes.c_ushort),
        ("dwFlags", ctypes.c_ulong),
        ("time", ctypes.c_ulong),
        ("dwExtraInfo", PUL)
    ]

class HardwareInput(ctypes.Structure):
    _fields_ = [
        ("uMsg", ctypes.c_ulong),
        ("wParamL", ctypes.c_short),
        ("wParamH", ctypes.c_ushort)
    ]

class MouseInput(ctypes.Structure):
    _fields_ = [
        ("dx", ctypes.c_long),
        ("dy", ctypes.c_long),
        ("mouseData", ctypes.c_ulong),
        ("dwFlags", ctypes.c_ulong),
        ("time",ctypes.c_ulong),
        ("dwExtraInfo", PUL)
    ]

class Input_I(ctypes.Union):
    _fields_ = [
        ("ki", KeyBdInput),
        ("mi", MouseInput),
        ("hi", HardwareInput)
    ]

class Input(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_ulong),
        ("ii", Input_I)
    ]

class Keyboard:
    """
    Class Keyboard
    :author: Paradoxis <luke@paradoxis.nl>
    :description:

    Keyboard methods to trigger fake key events
    """

    # Keyboard key constants
    # More information: https://msdn.microsoft.com/en-us/library/windows/desktop/dd375731(v=vs.85).aspx
    VK_BACKSPACE = 0x08
    VK_ENTER = 0x0D
    VK_CTRL = 0x11
    VK_ALT = 0x12
    VK_0 = 0x30
    VK_1 = 0x31
    VK_2 = 0x32
    VK_3 = 0x33
    VK_4 = 0x34
    VK_5 = 0x35
    VK_6 = 0x36
    VK_7 = 0x37
    VK_8 = 0x38
    VK_9 = 0x39
    VK_A = 0x41
    VK_B = 0x42
    VK_C = 0x43
    VK_D = 0x44
    VK_E = 0x45
    VK_F = 0x46
    VK_G = 0x47
    VK_H = 0x48
    VK_I = 0x49
    VK_J = 0x4A
    VK_K = 0x4B
    VK_L = 0x4C
    VK_M = 0x4D
    VK_N = 0x4E
    VK_O = 0x4F
    VK_P = 0x50
    VK_Q = 0x51
    VK_R = 0x52
    VK_S = 0x53
    VK_T = 0x54
    VK_U = 0x55
    VK_V = 0x56
    VK_W = 0x57
    VK_X = 0x58
    VK_Y = 0x59
    VK_Z = 0x5A
    VK_VOLUME_MUTE = 0xAD
    VK_VOLUME_DOWN = 0xAE
    VK_VOLUME_UP = 0xAF
    VK_MEDIA_NEXT_TRACK = 0xB0
    VK_MEDIA_PREV_TRACK = 0xB1
    VK_MEDIA_PLAY_PAUSE = 0xB3
    VK_MEDIA_STOP = 0xB2
    VK_LBUTTON = 0x01
    VK_RBUTTON = 0x02
    VK_CANCEL = 0x03
    VK_MBUTTON = 0x04
    VK_XBUTTON1 = 0x05
    VK_XBUTTON2 = 0x06
    VK_BACK = 0x08
    VK_TAB = 0x09
    VK_CLEAR = 0x0C
    VK_RETURN = 0x0D
    VK_SHIFT = 0x10
    VK_CONTROL = 0x11
    VK_MENU = 0x12
    VK_PAUSE = 0x13
    VK_CAPITAL = 0x14
    VK_KANA = 0x15
    VK_HANGUEL = 0x15
    VK_HANGUL = 0x15
    VK_JUNJA = 0x17
    VK_FINAL = 0x18
    VK_HANJA = 0x19
    VK_KANJI = 0x19
    VK_ESCAPE = 0x1B
    VK_CONVERT = 0x1C
    VK_NONCONVERT = 0x1D
    VK_ACCEPT = 0x1E
    VK_MODECHANGE = 0x1F
    VK_SPACE = 0x20
    VK_PRIOR = 0x21
    VK_NEXT = 0x22
    VK_END = 0x23
    VK_HOME = 0x24
    VK_LEFT = 0x25
    VK_UP = 0x26
    VK_RIGHT = 0x27
    VK_DOWN = 0x28
    VK_SELECT = 0x29
    VK_PRINT = 0x2A
    VK_EXECUTE = 0x2B
    VK_SNAPSHOT = 0x2C
    VK_INSERT = 0x2D
    VK_DELETE = 0x2E
    VK_HELP = 0x2F
    VK_LWIN = 0x5B
    VK_RWIN = 0x5C
    VK_APPS = 0x5D
    VK_SLEEP = 0x5F
    VK_NUMPAD0 = 0x60
    VK_NUMPAD1 = 0x61
    VK_NUMPAD2 = 0x62
    VK_NUMPAD3 = 0x63
    VK_NUMPAD4 = 0x64
    VK_NUMPAD5 = 0x65
    VK_NUMPAD6 = 0x66
    VK_NUMPAD7 = 0x67
    VK_NUMPAD8 = 0x68
    VK_NUMPAD9 = 0x69
    VK_MULTIPLY = 0x6A
    VK_ADD = 0x6B
    VK_SEPARATOR = 0x6C
    VK_SUBTRACT = 0x6D
    VK_DECIMAL = 0x6E
    VK_DIVIDE = 0x6F
    VK_F1 = 0x70
    VK_F2 = 0x71
    VK_F3 = 0x72
    VK_F4 = 0x73
    VK_F5 = 0x74
    VK_F6 = 0x75
    VK_F7 = 0x76
    VK_F8 = 0x77
    VK_F9 = 0x78
    VK_F10 = 0x79
    VK_F11 = 0x7A
    VK_F12 = 0x7B
    VK_F13 = 0x7C
    VK_F14 = 0x7D
    VK_F15 = 0x7E
    VK_F16 = 0x7F
    VK_F17 = 0x80
    VK_F18 = 0x81
    VK_F19 = 0x82
    VK_F20 = 0x83
    VK_F21 = 0x84
    VK_F22 = 0x85
    VK_F23 = 0x86
    VK_F24 = 0x87
    VK_NUMLOCK = 0x90
    VK_SCROLL = 0x91
    VK_LSHIFT = 0xA0
    VK_RSHIFT = 0xA1
    VK_LCONTROL = 0xA2
    VK_RCONTROL = 0xA3
    VK_LMENU = 0xA4
    VK_RMENU = 0xA5
    VK_BROWSER_BACK = 0xA6
    VK_BROWSER_FORWARD = 0xA7
    VK_BROWSER_REFRESH = 0xA8
    VK_BROWSER_STOP = 0xA9
    VK_BROWSER_SEARCH = 0xAA
    VK_BROWSER_FAVORITES = 0xAB
    VK_BROWSER_HOME = 0xAC
    VK_LAUNCH_MAIL = 0xB4
    VK_LAUNCH_MEDIA_SELECT = 0xB5
    VK_LAUNCH_APP1 = 0xB6
    VK_LAUNCH_APP2 = 0xB7
    VK_OEM_1 = 0xBA
    VK_OEM_PLUS = 0xBB
    VK_OEM_COMMA = 0xBC
    VK_OEM_MINUS = 0xBD
    VK_OEM_PERIOD = 0xBE
    VK_OEM_2 = 0xBF
    VK_OEM_3 = 0xC0
    VK_OEM_4 = 0xDB
    VK_OEM_5 = 0xDC
    VK_OEM_6 = 0xDD
    VK_OEM_7 = 0xDE
    VK_OEM_8 = 0xDF
    VK_OEM_102 = 0xE2
    VK_PROCESSKEY = 0xE5
    VK_PACKET = 0xE7
    VK_ATTN = 0xF6
    VK_CRSEL = 0xF7
    VK_EXSEL = 0xF8
    VK_EREOF = 0xF9
    VK_PLAY = 0xFA
    VK_ZOOM = 0xFB
    VK_NONAME = 0xFC
    VK_PA1 = 0xFD
    VK_OEM_CLEAR = 0xFE

    def keyDown(keyCode):
        """
        Key down wrapper
        :param keyCode: int
        :return: void
        """
        extra = ctypes.c_ulong(0)
        ii_ = Input_I()
        ii_.ki = KeyBdInput(keyCode, 0x48, 0, 0, ctypes.pointer(extra) )
        x = Input( ctypes.c_ulong(1), ii_ )
        SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    def keyUp(keyCode):
        """
        Key up wrapper
        :param keyCode: int
        :return: void
        """
        extra = ctypes.c_ulong(0)
        ii_ = Input_I()
        ii_.ki = KeyBdInput(keyCode, 0x48, 0x0002, 0, ctypes.pointer(extra) )
        x = Input( ctypes.c_ulong(1), ii_ )
        SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    def key(keyCode, length = 0):
        """
        Type a key
        :param keyCode: int
        :param length: int
        :return:
        """
        Keyboard.keyDown(keyCode)
        time.sleep(length)
        Keyboard.keyUp(keyCode)